/******************************************************************************
 *    HOST ERROR CODES
 ******************************************************************************
 *
 * Different return codes from the bootloader host.  Functions are not
 * limited to these values, but are encuraged to use them when returning
 * standard error values.
 *
 * 0 is successful, all other values indicate a failure.
 *****************************************************************************/
/* Completed successfully */
const CYRET_SUCCESS =           0x00;
/* File is not accessable */
const CYRET_ERR_FILE =          0x01;
/* Reached the end of the file */
const CYRET_ERR_EOF =           0x02;
/* The amount of data available is outside the expected range */
const CYRET_ERR_LENGTH =        0x03;
/* The data is not of the proper form */
const CYRET_ERR_DATA =          0x04;
/* The command is not recognized */
const CYRET_ERR_CMD =           0x05;
/* The expected device does not match the detected device */
const CYRET_ERR_DEVICE =        0x06;
/* The bootloader version detected is not supported */
const CYRET_ERR_VERSION =       0x07;
/* The checksum does not match the expected value */
const CYRET_ERR_CHECKSUM =      0x08;
/* The flash array is not valid */
const CYRET_ERR_ARRAY =         0x09;
/* The flash row is not valid */
const CYRET_ERR_ROW =           0x0A;
/* The bootloader is not ready to process data */
const CYRET_ERR_BTLDR =         0x0B;
/* The application is currently marked as active */
const CYRET_ERR_ACTIVE =        0x0C;
/* An unknown error occured */
const CYRET_ERR_UNK =           0x0F;
/* The operation was aborted */
const CYRET_ABORT =             0xFF;

/* The communications object reported an error */
const CYRET_ERR_COMM_MASK =     0x2000;
/* The bootloader reported an error */
const CYRET_ERR_BTLDR_MASK =    0x4000;


/******************************************************************************
 *    BOOTLOADER STATUS CODES
 ******************************************************************************
 *
 * Different return status codes from the bootloader.
 *
 * 0 is successful, all other values indicate a failure.
 *****************************************************************************/
/* Completed successfully */
const CYBTLDR_STAT_SUCCESS =       0x00;
/* The provided key does not match the expected value */
const CYBTLDR_STAT_ERR_KEY =       0x01;
/* The verification of flash failed */
const CYBTLDR_STAT_ERR_VERIFY =    0x02;
/* The amount of data available is outside the expected range */
const CYBTLDR_STAT_ERR_LENGTH =    0x03;
/* The data is not of the proper form */
const CYBTLDR_STAT_ERR_DATA =      0x04;
/* The command is not recognized */
const CYBTLDR_STAT_ERR_CMD =       0x05;
/* The expected device does not match the detected device */
const CYBTLDR_STAT_ERR_DEVICE =    0x06;
/* The bootloader version detected is not supported */
const CYBTLDR_STAT_ERR_VERSION =   0x07;
/* The checksum does not match the expected value */
const CYBTLDR_STAT_ERR_CHECKSUM =  0x08;
/* The flash array is not valid */
const CYBTLDR_STAT_ERR_ARRAY =     0x09;
/* The flash row is not valid */
const CYBTLDR_STAT_ERR_ROW =       0x0A;
/* The flash row is protected and can not be programmed */
const CYBTLDR_STAT_ERR_PROTECT =   0x0B;
/* The application is not valid and cannot be set as active */
const CYBTLDR_STAT_ERR_APP =       0x0C;
/* The application is currently marked as active */
const CYBTLDR_STAT_ERR_ACTIVE =    0x0D;
/* An unknown error occured */
const CYBTLDR_STAT_ERR_UNK =       0x0F;


/******************************************************************************
 *    VERSION INFORMATION
 ******************************************************************************
 *
 * Major � Used to indicate binary compatibility.  If a change is incompatible
 *         in any way with the prior release, the major version number will be
 *         updated.
 * Minor � Used to indicate feature set.  If a new feature or functionality is
 *         added beyond what was available in a prior release, the this number
 *         will be updated.
 * Patch � Used to indicate very minor fixes.  If the code was modified to fix
 *         a defect or to improve the quality in any way that does not add new
 *         functionality or change APIs this version number will be updated.
 *
 * 1.0   - Original (PSoC Creator 1.0 Beta 5)
 * 1.1   - Add checksum option (PSoC Creator 1.0 Production)
 * 1.2   - Add support for Multi Application Bootloaders
 *
 *****************************************************************************/
const VERSION_MAJOR =           0x01;
const VERSION_MINOR =           0x02;
const VERSION_PATCH =           0x00;

/* Maximum number of bytes to allocate for a single command.  */
const MAX_COMMAND_SIZE = 512;

/* Max XFER chunk size */
const MAX_XFER_SIZE = 64;

//STANDARD PACKET FORMAT:
// Multi byte entries are encoded in LittleEndian.
/*******************************************************************************
* [1-byte] [1-byte ] [2-byte] [n-byte] [ 2-byte ] [1-byte]
* [ SOP  ] [Command] [ Size ] [ Data ] [Checksum] [ EOP  ]
*******************************************************************************/


/* The first byte of any boot loader command. */
const CMD_START =               0x01;
/* The last byte of any boot loader command. */
const CMD_STOP =                0x17;
/* The minimum number of bytes in a bootloader command. */
const BASE_CMD_SIZE =           0x07;

/* Command identifier for verifying the checksum value of the bootloadable project. */
const CMD_VERIFY_CHECKSUM =     0x31;
/* Command identifier for getting the number of flash rows in the target device. */
const CMD_GET_FLASH_SIZE =      0x32;
/* Command identifier for getting info about the app status. This is only supported on multi app bootloader. */
const CMD_GET_APP_STATUS =      0x33;
/* Command identifier for reasing a row of flash data from the target device. */
const CMD_ERASE_ROW =           0x34;
/* Command identifier for making sure the bootloader host and bootloader are in sync. */
const CMD_SYNC =                0x35;
/* Command identifier for setting the active application. This is only supported on multi app bootloader. */
const CMD_SET_ACTIVE_APP =      0x36;
/* Command identifier for sending a block of data to the bootloader without doing anything with it yet. */
const CMD_SEND_DATA =           0x37;
/* Command identifier for starting the boot loader.  All other commands ignored until this is sent. */
const CMD_ENTER_BOOTLOADER =    0x38;
/* Command identifier for programming a single row of flash. */
const CMD_PROGRAM_ROW =         0x39;
/* Command identifier for verifying the contents of a single row of flash. */
const CMD_VERIFY_ROW =          0x3A;
/* Command identifier for exiting the bootloader and restarting the target program. */
const CMD_EXIT_BOOTLOADER =     0x3B;

/* Checksum type is a basic inverted summation of all bytes */
const SUM_CHECKSUM = 0x00;

/* 16-bit CRC checksum using the CCITT implementation */
const CRC_CHECKSUM = 0x01;

/* global state about fw file we opened 
*/

/* Variable used to store the currently selected packet checksum type */
var CyBtldr_Checksum = CRC_CHECKSUM;

function CyBtldr_ComputeChecksum(buf, size) {
    if (CyBtldr_Checksum == CRC_CHECKSUM) {
	    var crc = +0xffff;
	    var tmp;
	    var i, j;
           
	    if (size == 0)
		    return (~crc);
           
		j = +0;
	    do {
			//console.log('csb ' + i + ' + ' + j + ' is ' + buf[i + j]);
		    for (i = 0, tmp = +0x00ff & buf[j]; i < 8; i++, tmp >>= 1) {
			    if ((crc & +0x0001) ^ (tmp & +0x0001))
				    crc = (crc >> 1) ^ 0x8408;
			    else
			        crc >>= 1;
		    }
			j++;
	    } while (--size);

	    crc = ~crc;
        tmp = crc;
	    crc = (crc << 8) | ((tmp >> 8) & 0xFF);  
	    return crc;
    }
	else { /* SUM_CHECKSUM */
        var sum = 0;

		for (i = 0; i < size; i++)
		    sum += buf[i];
	    return (1 + ~sum);
    }
}

function CyBtldr_SetCheckSumType(chksumType) {
    CyBtldr_Checksum = chksumType;
}

function CyBtldr_SendCommand(connId, cmdArray, replyCallback) {
	hidWriteData(connId, cmdArray, 0, replyCallback);
}

function CyBtldr_ReadReply(connId, replyCallback) {
	hidReadData(connId, replyCallback);
}

var bootLoaderState = 0;

const BOOTLS_INIT		= 0;
const BOOTLS_ROW		= 1;
const BOOTLS_SENDROW	= 2;
const BOOTLS_END		= 3;
const BOOTLS_DONE		= 4;
const BOOTLS_ERREND		= 100;
const BOOTLS_ERROR		= 200;

var cyStatReport;
var cyProgReport = null;
var cyDoneReport = null;

function bootLoaderFail(connId, msg) {
	cyStatReport(msg);
	// if started op, finish op
	if (bootLoaderState > 0)
		bootLoaderState = BOOTLS_ERREND;
	else
		bootLoaderState = BOOTLS_ERROR;
	setTimeout(function() {			
			bootLoad(connId, 'error');
		}, 0);
}

function bootLoaderAdvance(connId, msg) {
	// schedule the next operation in the run loop
	// instead of recursing, to avoid stack getting too deep
	setTimeout(function() {
			if (bootLoaderState == 0)
				console.log('bad state');
			bootLoad(connId, msg);
		}, 0);
}

var hdrSiliconId    = +0;
var hdrSiliconRev   = +0;
var hdrChecksumType = +0;

var cyFileBytes = null;
var cyFileLength = +0;
var cyFileOffset = +0;

function CyBtldr_Readline() {
	if (cyFileBytes == null)
		return null;
	if (cyFileOffset >= cyFileBytes)
		return null;
	var gotCR = false;
	for (var x = cyFileOffset; x < cyFileLength; x++) {
		if (cyFileBytes[x] == 0xD)
			gotCR = true;
		if (cyFileBytes[x] == 0xA)
			break;
	}
	if (x <= cyFileOffset)
		return null;
	var strLen = x - cyFileOffset + 1;
	var cpyLen = strLen - (gotCR ? 2 : 1);
	var retBufArray = new ArrayBuffer(cpyLen);
	var retBuf = new Uint8Array(retBufArray);

	for (var y = 0; y < cpyLen; y++) {
		retBuf[y] = cyFileBytes[cyFileOffset + y];
	}
	cyFileOffset += strLen;
	return retBufArray;
}

function CyBtldr_HexFromASCII(asciiArray, astart, alength) {
	var hexArray = new ArrayBuffer(alength);
	var hexView  = new Uint8Array(hexArray);
	var ascView  = new Uint8Array(asciiArray);
	var ca = +0, cb = +0;

	for (var ci = 0; ci < alength; ci++) {
		ca = ascView[astart + ci * 2];
		cb = ascView[astart + ci * 2 + 1];
		
		if (ca >= 0x61 && ca <= 0x66)
			ca = ca - 0x61 + 10;
		else if (ca >= 0x41 && ca <= 0x46)
			ca = ca - 0x41 + 10;
		else if (ca >= 0x30 && ca <= 0x39)
			ca = ca - 0x30;
		else
			return null;

		if (cb >= 0x61 && cb <= 0x66)
			cb = cb - 0x61 + 10;
		else if (cb >= 0x41 && cb <= 0x46)
			cb = cb - 0x41 + 10;
		else if (cb >= 0x30 && cb <= 0x39)
			cb = cb - 0x30;
		else
			return null;

		hexView[ci] = (ca << 4) | cb;
	}
	return hexArray;
}

// row sending incremental state
//
var rowOffset  = 0;
var rowLength  = 0;
var rowArrayId = 0;
var rowNum     = 0;
var rowSize    = 0; // includes all bytes
var rowLength  = 0; // only data bytes
var rowOffset  = 0;
var rowSize    = 0;
var rowSum     = 0;
var rowBytes;

function CyBtldr_VerifyRow(connId) {
	bootLoaderAdvance(connId, 'programmed');
}

function CyBtldr_ParseProgramRowResult(connId, laststat, reply) {
	if (laststat == false) {
		bootLoaderFail('Failed to program row');
		return;
	}
	var cmdBuf = new Uint8Array(reply);
	if (cmdBuf[0] != CMD_START) {
		bootLoaderFail(connId, 'ProgRow cmd[0] not START');
		return;
	}
	if (cmdBuf[1] != CYRET_SUCCESS) {
		bootLoaderFail(connId, 'ProgRow cmd[1] not SUCCESS (' + cmdBuf[1] + ')');
		return;
	}
	CyBtldr_VerifyRow(connId);
}

function CyBtldr_ProgramRow(connId) {
	// create a program row command, limiting payload such that we 
	// stay under our max transport chunk
	//	
	const PROGROW_DATA_SIZE = 3;
	var rowChunk = rowLength - rowOffset;
	if ((rowChunk + (BASE_CMD_SIZE + PROGROW_DATA_SIZE)) > MAX_XFER_SIZE) {
		bootLoaderFail('ProgRow too big');
		return;
	}
	var hidBaseSize = BASE_CMD_SIZE + PROGROW_DATA_SIZE + rowChunk;
	var hidSize  = ((((hidBaseSize + 63) / 64) | 0) * 64); // round up 64 byte chunk for usb
	var cmdArray = new ArrayBuffer(hidSize);
	var cmdBuf   = new Uint8Array(cmdArray);

    cmdBuf[0] = CMD_START;
    cmdBuf[1] = CMD_PROGRAM_ROW;
    cmdBuf[2] = (rowChunk + PROGROW_DATA_SIZE) & 0xFF;
    cmdBuf[3] = ((rowChunk + PROGROW_DATA_SIZE) >> 8) & 0xFF;
    cmdBuf[4] = rowArrayId;
    cmdBuf[5] = rowNum & 0xFF;
    cmdBuf[6] = (rowNum >> 8) & 0xFF;
    for (var i = 0; i < rowChunk; i++)
        cmdBuf[i + 7] = rowBytes[rowOffset + 5 + i];
	var csoff = i + 7;
    checksum = CyBtldr_ComputeChecksum(cmdBuf, csoff);
    cmdBuf[csoff++] = checksum & 0xFF;
    cmdBuf[csoff++] = (checksum >> 8) & 0xFF;
    cmdBuf[csoff++] = CMD_STOP;
	if (csoff != hidBaseSize) {
		bootLoaderFail(connId, '-Bad progrow Cmd');
		return;
	}
	rowOffset += rowChunk;
	
	//cyStatReport('PROG Row num=' + rowNum + ' chunk=' + rowChunk + ' nxtoff=' + rowOffset);
	// here goes, program the row
	// send it, parsing the results in callback
	CyBtldr_SendCommand(connId, cmdArray, function(connId, laststat, reportId) {
		// parse response
		CyBtldr_ReadReply(connId, CyBtldr_ParseProgramRowResult);
	});
}

function CyBtldr_ParseSendRowResult(connId, laststat, reply) {
	if (laststat == false) {
		bootLoaderFail('Failed to send row');
		return;
	}
	var cmdBuf = new Uint8Array(reply);
	if (cmdBuf[0] != CMD_START) {
		bootLoaderFail(connId, 'SendRow cmd[0] not START');
		return;
	}
	if (cmdBuf[1] != CYRET_SUCCESS) {
		bootLoaderFail(connId, 'SendRow cmd[1] not SUCCESS (' + cmdBuf[1] + ')');
		return;
	}
	bootLoaderAdvance(connId, 'sent chunk');
}

function CyBtldr_SendRowData(connId) {
	if (rowOffset >= rowLength) {
		// complete
		return;
	}
	// create a send row command, limiting payload such that we 
	// stay under our max transport chunk
	//	
	var rowChunk = rowLength - rowOffset;
	if ((rowChunk + BASE_CMD_SIZE) > MAX_XFER_SIZE)
		rowChunk = MAX_XFER_SIZE - BASE_CMD_SIZE;
	var hidBaseSize = BASE_CMD_SIZE + rowChunk;
	var hidSize  = ((((hidBaseSize + 63) / 64) | 0) * 64); // round up 64 byte chunk for usb
	var cmdArray = new ArrayBuffer(hidSize);
	var cmdBuf   = new Uint8Array(cmdArray);

    cmdBuf[0] = CMD_START;
    cmdBuf[1] = CMD_SEND_DATA;
    cmdBuf[2] = (rowChunk) & 0xFF;
    cmdBuf[3] = ((rowChunk) >> 8) & 0xFF;
    for (var i = 0; i < rowChunk; i++)
        cmdBuf[i + 4] = rowBytes[rowOffset + 5 + i];
	var csoff = i + 4;
    checksum = CyBtldr_ComputeChecksum(cmdBuf, csoff);
    cmdBuf[csoff++] = checksum & 0xFF;
    cmdBuf[csoff++] = (checksum >> 8) & 0xFF;
    cmdBuf[csoff++] = CMD_STOP;
	if (csoff != hidBaseSize) {
		bootLoaderFail(connId, '-Bad sendrow Cmd');
		return;
	}
	rowOffset += rowChunk;

	//cyStatReport('Send Row num=' + rowNum + ' chunk=' + rowChunk + ' nxtoff=' + rowOffset);
	// here goes, send the data
	//
	CyBtldr_SendCommand(connId, cmdArray, function(connId, laststat, reportId) {
		// parse response
		CyBtldr_ReadReply(connId, CyBtldr_ParseSendRowResult);
	});
}

function CyBtldr_ParseRowData(connId) {
	if (cyFileOffset < (cyFileLength - 7)) {
		var row = CyBtldr_Readline();
		
		if (row == null) {
			bootLoaderFail(connId, '-No Row');
			return;
		}
		if (row.byteLength < 12) {
			bootLoaderFail(connId, '-No full Row');
			return;
		}
		// Format of Row is ":" 1-byte-id, 2-byte-address, 2-byte-size, <size bytes>, 1-byte-checksum
		//
		var rowhex = CyBtldr_HexFromASCII(row, 1, ((row.byteLength + 1) / 2 - 1) | 0);
		if (rowhex == null) {
			bootLoaderFail(connId, '-Bad Row');
			return;
		}
		// setup incremental state of row
		//
		rowBytes = new Uint8Array(rowhex);
		rowSize  = rowBytes.length;
		
		rowAarrayId = rowBytes[0];
		rowNum      = (rowBytes[1] << 8) | rowBytes[2];
		rowLength   = (rowBytes[3] << 8) | rowBytes[4];
		rowSum      = rowBytes[rowLength - 1];
		rowOffset   = 0;

		//cyStatReport('Row num=' + rowNum + ' size=' + rowSize + ' sum=' + rowSum);
	} else {
		rowLength = 0;
		rowOffset = 0;
	}
	bootLoaderAdvance(connId, 'parsed row');
}

function CyBtldr_ParseExitBootLoaderCmdResult(connId, laststat, reply) {
	if (laststat == false) {
		bootLoaderFail('Failed to exit bootloader');
		return;
	}
	var cmdBuf = new Uint8Array(reply);
	if (cmdBuf[0] != CMD_START) {
		bootLoaderFail(connId, 'cmd[0] not START');
		return;
	}
	if (cmdBuf[1] != CYRET_SUCCESS) {
		bootLoaderFail(connId, 'cmd[1] not SUCCESS (' + cmdBuf[1] + ')');
		return;
	}
	bootLoaderFail('Failed to exit bootloader');
}

function CyBtldr_EndBootloadOperation(connId) {

	// create an exit-boot-load command
	var cmdArray = new ArrayBuffer(64/*BASE_CMD_SIZE*/);
	var cmdBuf = new Uint8Array(cmdArray);

	var resetType = +0; // RESET
		
	cyStatReport('Exiting bootload mode');

	cmdBuf[0] = CMD_START;
    cmdBuf[1] = CMD_EXIT_BOOTLOADER;
    cmdBuf[2] = 1;
    cmdBuf[3] = 0;
    cmdBuf[4] = resetType;
    var checksum = CyBtldr_ComputeChecksum(cmdBuf, BASE_CMD_SIZE + 1 - 3);
    cmdBuf[5] = +checksum & 0xFF;
    cmdBuf[6] = (+checksum >> 8) & 0xFF;
    cmdBuf[7] = CMD_STOP;
	
	// send the exit command. if all went well, this will reset/reboot the
	// device, which means there will be NO response and getting one means
	// the download actually failed
	//
	CyBtldr_SendCommand(connId, cmdArray, function(connId, laststat, reportId) {
		cyStatReport('Wait for system reset');		
	});
	//cyProgReport(100);
	bootLoaderAdvance(connId, 'ended');
}

function CyBtldr_ParseEnterBootLoaderCmdResult(connId, laststat, reply) {
	if (laststat == false) {
		bootLoaderFail('Failed to enter bootloader');
		return;
	}
	var cmdBuf = new Uint8Array(reply);
	if (cmdBuf[0] != CMD_START) {
		bootLoaderFail(connId, 'cmd[0] not START');
		return;
	}
	if (cmdBuf[1] != CYRET_SUCCESS) {
		bootLoaderFail(connId, 'cmd[1] not SUCCESS');
		return;
	}
	if (cmdBuf[2] != 8) {
		bootLoaderFail(connId, 'cmd[2] not 8');
		return;
	}
	if (cmdBuf[3] != 0) {
		bootLoaderFail(connId, 'cmd[3] not 0');
		return;
	}
    var siliconId = (cmdBuf[7] << 24) | (cmdBuf[6] << 16) | (cmdBuf[5] << 8) | cmdBuf[4];
    var siliconRev = cmdBuf[8];
    var blVersion = (cmdBuf[11] << 16) | (cmdBuf[10] << 8) | cmdBuf[9];

	console.log('siliconId =' + siliconId);
	console.log('siliconRev=' + siliconRev);
	console.log('blVersion =' + blVersion);

	//cyStatReport('Bootloader found, rev=' + blVersion + ' starting upgrade');
  cyStatReport('Programming bootloader...');
  
	if (cyFileBytes != null) {
		// make sure silicon id/rev matches
		//
		if (siliconId != hdrSiliconId) {
			bootLoaderFail(connId, 'Silicon ID mismatch' + ' hdr='+hdrSiliconId);
			return;
		}
		if (siliconRev != hdrSiliconRev) {
			bootLoaderFail(connId, 'Silicon Rev mismatch');
			return;
		}
	} else {
		// no file, i.e. testing, so go to end state now
		bootLoaderState = BOOTLS_END;
	}	
	// next state please
	bootLoaderAdvance(connId, 'started');
}
	
function CyBtldr_StartBootloadOperation(connId) {

	// create an enter-boot-load command
	var cmdArray = new ArrayBuffer(64/*BASE_CMD_SIZE*/);
	var cmdBuf = new Uint8Array(cmdArray);
	
	cyStatReport('Entering bootload mode');

    cmdBuf[0] = CMD_START;
    cmdBuf[1] = CMD_ENTER_BOOTLOADER;
    cmdBuf[2] = 0;
    cmdBuf[3] = 0;
    var checksum = CyBtldr_ComputeChecksum(cmdBuf, BASE_CMD_SIZE - 3);
    cmdBuf[4] = +checksum & 0xFF;
    cmdBuf[5] = (+checksum >> 8) & 0xFF;
    cmdBuf[6] = CMD_STOP;
	
	// send it, parsing the results in callback
	CyBtldr_SendCommand(connId, cmdArray, function(connId, laststat, reportId) {
		// parse response
		CyBtldr_ReadReply(connId, CyBtldr_ParseEnterBootLoaderCmdResult);
	});
}

function bootLoad(connId, opcode) {
	if (connId < 0) {
		console.log('BOOTLDR: No connection');
		cyStatReport('BOOTLDR: No connection');

		return;
	}
	console.log('BOOTLDR st:' + bootLoaderState + ' op:' + opcode);
 
	switch (bootLoaderState) {
	case BOOTLS_INIT:
		bootLoaderState = BOOTLS_ROW;
		CyBtldr_StartBootloadOperation(connId);
		break;
	case BOOTLS_ROW:
		if (cyFileOffset < (cyFileLength - 7)) {
			// setup to send one row of data
			bootLoaderState = BOOTLS_SENDROW;
			CyBtldr_ParseRowData(connId);
			if (cyProgReport !== null)
				cyProgReport(100 * (cyFileOffset / cyFileLength));
		}
		else {
			// did all rows, end
			bootLoaderState = BOOTLS_END;
			bootLoaderAdvance(connId, 'done programming');
		}
		break;
	case BOOTLS_SENDROW:
		if (rowOffset < rowLength) {
			if (((rowLength - rowOffset) + 10) <= MAX_XFER_SIZE) {
				// last chunk, program it
				//
				bootLoaderState = BOOTLS_ROW;
				CyBtldr_ProgramRow(connId);
			}
			else {
				// keep sending row data
				bootLoaderState = BOOTLS_SENDROW;
				CyBtldr_SendRowData(connId);
			}
		} else {
			bootLoaderState = BOOTLS_ROW;
			bootLoaderAdvance('This Cant happen');
		}
		break;
	case BOOTLS_ERREND:
	case BOOTLS_END:
		bootLoaderState = (bootLoaderState == BOOTLS_ERREND) ? BOOTLS_ERROR : BOOTLS_DONE;
		CyBtldr_EndBootloadOperation(connId);
		break;
	case BOOTLS_DONE:
	case BOOTLS_ERROR:
	default:
		var finOK = (bootLoaderState == BOOTLS_DONE) ? true : false;
		if (finOK)
			cyStatReport('USB Firmware Complete');
		else if (upgradeAborted)
			cyStatReport('USB Firmware Aborted');
		else
			cyStatReport('USB Firmware - Errors');
		hidCloseDevice(connId);
		if (cyDoneReport !== null)
			cyDoneReport(finOK);
		break;
	}
}

function startBootLoader() {
	if (cyFileBytes !== null) {
		// parse out the header from the image
		//
		var hdr = CyBtldr_Readline();

		if (hdr === null) {
			bootLoaderFail(null, '-No Header');
			return;
		}
		if (hdr.byteLength < 12) {
			bootLoaderFail(null, '-No full Header');
			return;
		}
		var hdrhex = CyBtldr_HexFromASCII(hdr, 0, 6);
		if (hdrhex === null) {
			bootLoaderFail(null, '-Bad Header');
			return;
		}
		var hdrbytes = new Uint8Array(hdrhex);	
		
		// first 4 bytes are siliconID
		//
		hdrSiliconId    = (hdrbytes[0] << 24) |
				(hdrbytes[1] << 16) | (hdrbytes[2] << 8) | hdrbytes[3];
		hdrSiliconRev   = hdrbytes[4];
		hdrChecksumType = hdrbytes[5];
		/*
		cyStatReport('HDR: silid=' + hdrSiliconId + ' rev= ' + hdrSiliconRev + 
				' cs=' + hdrChecksumType);
		*/
		cyStatReport('Opened Firmware OK');

		// set checkum type to header type
		//
		CyBtldr_SetCheckSumType(hdrChecksumType);
	}
	else {
		// no file, must be testing
		//
		CyBtldr_SetCheckSumType(CRC_CHECKSUM);
	}
	// the device should have re-enumerarted as cy-device so
	// find that and do the i/o from there. note because of
	// that, the serial port went away and rxdata is probably
	// garbage so don't use any more serial, or look at rxdata
	//

	bootLoaderState = BOOTLS_INIT;
	checkDevice(hidDevUpdate);
}

function hidDevUpdate(connected) {
		while (connected.state != 'Bootloader') {
		  cyStatReport('waiting for bootloader...');
		  setTimeout(function() { }, 100);
    }

  	hidUseDevice(CYP_VENDOR_ID, CYP_PRODUCT_ID, bootLoad, 'hello');
}

function startBootLoaderWait(laststat, rxdata) {
	if (laststat === false) {
		reportStatus('bootmode command failed');
		abortUpgrade();
		return;
	}
	// wait for re-enumeration
	//setTimeout(function() {
	//		startBootLoader();
	//	},
		//5500
	//);
}

function CyBtldr_Upgrade(filecontents, statusUpdater, progressUpdater, onFinished) {
	cyFileBytes = new Uint8Array(filecontents);
	cyFileOffset = 0;
	cyFileLength = cyFileBytes.length;
	cyStatReport = statusUpdater;
	cyProgReport = progressUpdater;
	cyDoneReport = onFinished;
	
	// if in boot-mode already, go do it, else switch to bootmode
	checkDevice(cybootDev);
}
	  
function cybootDev(connected) {
	  
		switch (connected.type) {
		case USBDEV_NONE:
			console.log('not connected');
			break;
		case USBDEV_REVO1:
		case USBDEV_REVO2:
		default:
			console.log('not in bootloader mode');
			// set bootloader mode
			//runCommand('cyboot', 'ruby', null, false, startBootLoaderWait);
			break;
		case USBDEV_CYP:
			startBootLoader();
			break;
		}
}

function stubStatus(status) {
	console.log(status);
}

function stubProgress(progress) {
}

function stubFinished(laststat) {
}

function CyBtldr_Test() {
	// open file and read contents
	var fname = './USB_Digital_Audio.cyacd'; //'./cypressUSB.cyacd';
    var rawFile = new XMLHttpRequest();

	// get test file data into an Array Buffer
	//
    rawFile.open("GET", fname, true);
    rawFile.onreadystatechange = function () {
        if(rawFile.readyState === 4) {
            if(rawFile.status === 200 || rawFile.status == 0)  {
				if (rawFile.responseText != null) {
	                var fileText = rawFile.responseText;
					var buf = new ArrayBuffer(fileText.length);
					var bufView = new Uint8Array(buf);
					for (var i = 0, strLen = fileText.length; i < strLen; i++) {
						bufView[i] = fileText.charCodeAt(i);
					}
					CyBtldr_Upgrade(buf, stubStatus, stubProgress, stubFinished);
				}
				else {
					console.log('no file result');
				}
            }
        }
    }
    rawFile.send(null);
}

